{% extends "base.html" %}
{% block extrahead %}
<meta name="google-adsense-account" content="ca-pub-2802331499006697">
<script async src="https://pagead2.googlesyndication.com/pagead/js/adsbygoogle.js?client=ca-pub-2802331499006697"
        crossorigin="anonymous"></script>
{% endblock %}